"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const js_yaml_1 = __importDefault(require("js-yaml"));
const glob_to_regexp_1 = __importDefault(require("glob-to-regexp"));
const binary_extensions_1 = __importDefault(require("binary-extensions"));
const isbinaryfile_1 = require("isbinaryfile");
const walk_tree_1 = __importDefault(require("./helpers/walk-tree"));
const load_data_1 = __importDefault(require("./helpers/load-data"));
const read_file_1 = __importDefault(require("./helpers/read-file"));
const convert_pcre_1 = __importDefault(require("./helpers/convert-pcre"));
const convert_glob_1 = __importDefault(require("./helpers/convert-glob"));
async function analyse(input, opts = {}) {
    var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q, _r;
    var _s, _t, _u;
    const useRawContent = opts.fileContent !== undefined;
    const langData = await (0, load_data_1.default)('languages.yml').then(js_yaml_1.default.load);
    const vendorData = await (0, load_data_1.default)('vendor.yml').then(js_yaml_1.default.load);
    const heuristicsData = await (0, load_data_1.default)('heuristics.yml').then(js_yaml_1.default.load);
    const generatedData = await (0, load_data_1.default)('generated.rb').then(text => { var _a; return (_a = text.match(/(?<=name\.match\(\/).+?(?=(?<!\\)\/\))/gm)) !== null && _a !== void 0 ? _a : []; });
    vendorData.push(...generatedData);
    const fileAssociations = {};
    const extensions = {};
    const overrides = {};
    const results = {
        files: { count: 0, bytes: 0, results: {} },
        languages: { count: 0, bytes: 0, results: {} },
        unknown: { count: 0, bytes: 0, extensions: {}, filenames: {} },
    };
    const ignoredFiles = [
        /\/\.git\//,
        opts.keepVendored ? [] : vendorData.map(path => RegExp(path)),
        (_b = (_a = opts.ignoredFiles) === null || _a === void 0 ? void 0 : _a.map(path => (0, glob_to_regexp_1.default)('*' + path + '*', { extended: true }))) !== null && _b !== void 0 ? _b : [],
    ].flat();
    let files, folders;
    if (opts.fileContent) {
        opts.fileContent = Array.isArray(opts.fileContent) ? opts.fileContent : [opts.fileContent];
        files = [`${input}`];
        folders = [''];
    }
    else {
        const data = (0, walk_tree_1.default)(input !== null && input !== void 0 ? input : '.', ignoredFiles);
        ({ files, folders } = data);
    }
    // Apply aliases
    opts = { checkIgnored: !opts.quick, checkAttributes: !opts.quick, checkHeuristics: !opts.quick, checkShebang: !opts.quick, ...opts };
    // Ignore specific languages
    for (const lang of (_c = opts.ignoredLanguages) !== null && _c !== void 0 ? _c : []) {
        for (const key in langData) {
            if (lang.toLowerCase() === key.toLowerCase()) {
                delete langData[key];
                break;
            }
        }
    }
    // Load gitattributes
    const customIgnored = [];
    const customBinary = [];
    const customText = [];
    if (!useRawContent && !opts.quick) {
        for (const folder of folders) {
            // Skip if folder is marked in gitattributes
            if (customIgnored.some(path => (0, convert_glob_1.default)(path).test(folder)))
                continue;
            // Parse gitignores
            const ignoresFile = path_1.default.join(folder, '.gitignore');
            if (opts.checkIgnored && fs_1.default.existsSync(ignoresFile)) {
                const ignoresData = await (0, read_file_1.default)(ignoresFile);
                const ignoresList = ignoresData.split(/\r?\n/).filter(line => line.trim() && !line.startsWith('#'));
                const ignoredPaths = ignoresList.map(path => (0, convert_glob_1.default)(path).source);
                customIgnored.push(...ignoredPaths);
            }
            // Parse gitattributes
            const attributesFile = path_1.default.join(folder, '.gitattributes');
            if (opts.checkAttributes && fs_1.default.existsSync(attributesFile)) {
                const attributesData = await (0, read_file_1.default)(attributesFile);
                const relPathToRegex = (path) => (0, convert_glob_1.default)(path).source.substr(1).replace(folder, '');
                // Explicit text/binary associations
                const contentTypeMatches = attributesData.matchAll(/^(\S+).*?(-?binary|-?text)(?!=auto)/gm);
                for (const [_line, path, type] of contentTypeMatches) {
                    if (['text', '-binary'].includes(type))
                        customText.push(relPathToRegex(path));
                    if (['-text', 'binary'].includes(type))
                        customBinary.push(relPathToRegex(path));
                }
                // Custom vendor options
                const vendorMatches = attributesData.matchAll(/^(\S+).*[^-]linguist-(vendored|generated|documentation)(?!=false)/gm);
                for (const [_line, path] of vendorMatches) {
                    customIgnored.push(relPathToRegex(path));
                }
                // Custom file associations
                const customLangMatches = attributesData.matchAll(/^(\S+).*[^-]linguist-language=(\S+)/gm);
                for (let [_line, path, forcedLang] of customLangMatches) {
                    // If specified language is an alias, associate it with its full name
                    if (!langData[forcedLang]) {
                        const overrideLang = Object.entries(langData).find(entry => { var _a; return (_a = entry[1].aliases) === null || _a === void 0 ? void 0 : _a.includes(forcedLang.toLowerCase()); });
                        if (overrideLang)
                            forcedLang = overrideLang[0];
                    }
                    const fullPath = folder + (0, convert_glob_1.default)(path).source.substr(1);
                    overrides[fullPath] = forcedLang;
                }
            }
        }
    }
    // Check vendored files
    if (!useRawContent && !opts.keepVendored) {
        // Filter out any files that match a vendor file path
        const matcher = (match) => RegExp(match.replace(/\/$/, '/.+$').replace(/^\.\//, ''));
        files = files.filter(file => !customIgnored.some(pattern => matcher(pattern).test(file)));
    }
    // Load all files and parse languages
    const addResult = (file, result) => {
        if (!fileAssociations[file]) {
            fileAssociations[file] = [];
            extensions[file] = '';
        }
        const parent = !opts.childLanguages && result && langData[result].group || false;
        fileAssociations[file].push(parent || result);
        extensions[file] = path_1.default.extname(file).toLowerCase();
    };
    const overridesArray = Object.entries(overrides);
    // List all languages that could be associated with a given file
    for (const file of files) {
        let firstLine;
        if (opts.fileContent) {
            firstLine = (_f = (_e = (_d = opts.fileContent) === null || _d === void 0 ? void 0 : _d[files.indexOf(file)]) === null || _e === void 0 ? void 0 : _e.split('\n')[0]) !== null && _f !== void 0 ? _f : null;
        }
        else {
            if (!fs_1.default.existsSync(file) || fs_1.default.lstatSync(file).isDirectory())
                continue;
            firstLine = await (0, read_file_1.default)(file, true).catch(() => null);
        }
        // Skip if file is unreadable
        if (firstLine === null)
            continue;
        // Check shebang line for explicit classification
        if (!opts.quick && opts.checkShebang && firstLine.startsWith('#!')) {
            // Find matching interpreters
            const matches = Object.entries(langData).filter(([, data]) => { var _a; return (_a = data.interpreters) === null || _a === void 0 ? void 0 : _a.some(interpreter => firstLine.match('\\b' + interpreter + '\\b')); });
            if (matches.length) {
                // Add explicitly-identified language
                const forcedLang = matches[0][0];
                addResult(file, forcedLang);
            }
        }
        // Check override for manual language classification
        if (!useRawContent && !opts.quick && opts.checkAttributes) {
            const match = overridesArray.find(item => RegExp(item[0]).test(file));
            if (match) {
                const forcedLang = match[1];
                addResult(file, forcedLang);
                continue;
            }
        }
        // Search each language
        let skipExts = false;
        for (const lang in langData) {
            // Check if filename is a match
            const matchesName = (_g = langData[lang].filenames) === null || _g === void 0 ? void 0 : _g.some(name => path_1.default.basename(file.toLowerCase()) === name.toLowerCase());
            if (matchesName) {
                addResult(file, lang);
                skipExts = true;
            }
        }
        if (!skipExts)
            for (const lang in langData) {
                // Check if extension is a match
                const matchesExt = (_h = langData[lang].extensions) === null || _h === void 0 ? void 0 : _h.some(ext => file.toLowerCase().endsWith(ext.toLowerCase()));
                if (matchesExt)
                    addResult(file, lang);
            }
        // Fallback to null if no language matches
        if (!fileAssociations[file])
            addResult(file, null);
    }
    // Narrow down file associations to the best fit
    for (const file in fileAssociations) {
        // Skip binary files
        if (!useRawContent && !opts.keepBinary) {
            const isCustomText = customText.some(path => RegExp(path).test(file));
            const isCustomBinary = customBinary.some(path => RegExp(path).test(file));
            const isBinaryExt = binary_extensions_1.default.some(ext => file.endsWith('.' + ext));
            if (!isCustomText && (isCustomBinary || isBinaryExt || await (0, isbinaryfile_1.isBinaryFile)(file))) {
                continue;
            }
        }
        // Parse heuristics if applicable
        if (opts.checkHeuristics)
            for (const heuristics of heuristicsData.disambiguations) {
                // Make sure the extension matches the current file
                if (!heuristics.extensions.includes(extensions[file])) {
                    continue;
                }
                // Load heuristic rules
                for (const heuristic of heuristics.rules) {
                    // Make sure the language is not an array
                    if (Array.isArray(heuristic.language)) {
                        heuristic.language = heuristic.language[0];
                    }
                    // Make sure the results includes this language
                    const languageGroup = (_j = langData[heuristic.language]) === null || _j === void 0 ? void 0 : _j.group;
                    const matchesLang = fileAssociations[file].includes(heuristic.language);
                    const matchesParent = languageGroup && fileAssociations[file].includes(languageGroup);
                    if (!matchesLang && !matchesParent)
                        continue;
                    // Normalise heuristic data
                    const patterns = [];
                    const normalise = (contents) => patterns.push(...(Array.isArray(contents) ? contents : [contents]));
                    if (heuristic.pattern)
                        normalise(heuristic.pattern);
                    if (heuristic.named_pattern)
                        normalise(heuristicsData.named_patterns[heuristic.named_pattern]);
                    // Check file contents and apply heuristic patterns
                    const fileContent = await (0, read_file_1.default)(file).catch(() => null);
                    if (fileContent === null)
                        continue;
                    if (!patterns.length || patterns.some(pattern => (0, convert_pcre_1.default)(pattern).test(fileContent))) {
                        results.files.results[file] = heuristic.language;
                        break;
                    }
                }
            }
        // If no heuristics, assign a language
        (_k = (_s = results.files.results)[file]) !== null && _k !== void 0 ? _k : (_s[file] = fileAssociations[file][0]);
    }
    // Skip specified categories
    if ((_l = opts.categories) === null || _l === void 0 ? void 0 : _l.length) {
        const categories = ['data', 'markup', 'programming', 'prose'];
        const hiddenCategories = categories.filter(cat => !opts.categories.includes(cat));
        for (const [file, lang] of Object.entries(results.files.results)) {
            if (!hiddenCategories.some(cat => { var _a; return lang && ((_a = langData[lang]) === null || _a === void 0 ? void 0 : _a.type) === cat; }))
                continue;
            delete results.files.results[file];
            if (lang)
                delete results.languages.results[lang];
        }
        for (const category of hiddenCategories) {
            for (const [lang, { type }] of Object.entries(results.languages.results)) {
                if (type === category)
                    delete results.languages.results[lang];
            }
        }
    }
    // Convert paths to relative
    if (!useRawContent && opts.relativePaths) {
        const newMap = {};
        for (const [file, lang] of Object.entries(results.files.results)) {
            let relPath = path_1.default.relative(process.cwd(), file).replace(/\\/g, '/');
            if (!relPath.startsWith('../'))
                relPath = './' + relPath;
            newMap[relPath] = lang;
        }
        results.files.results = newMap;
    }
    // Load language bytes size
    for (const [file, lang] of Object.entries(results.files.results)) {
        if (lang && !langData[lang])
            continue;
        const fileSize = (_p = (_o = (_m = opts.fileContent) === null || _m === void 0 ? void 0 : _m[files.indexOf(file)]) === null || _o === void 0 ? void 0 : _o.length) !== null && _p !== void 0 ? _p : fs_1.default.statSync(file).size;
        results.files.bytes += fileSize;
        // If no language found, add extension in other section
        if (!lang) {
            const ext = path_1.default.extname(file);
            const unknownType = ext === '' ? 'filenames' : 'extensions';
            const name = ext === '' ? path_1.default.basename(file) : ext;
            (_q = (_t = results.unknown[unknownType])[name]) !== null && _q !== void 0 ? _q : (_t[name] = 0);
            results.unknown[unknownType][name] += fileSize;
            results.unknown.bytes += fileSize;
            continue;
        }
        // Add language and bytes data to corresponding section
        const { type } = langData[lang];
        (_r = (_u = results.languages.results)[lang]) !== null && _r !== void 0 ? _r : (_u[lang] = { type, bytes: 0, color: langData[lang].color });
        if (opts.childLanguages)
            results.languages.results[lang].parent = langData[lang].group;
        results.languages.results[lang].bytes += fileSize;
        results.languages.bytes += fileSize;
    }
    // Set counts
    results.files.count = Object.keys(results.files.results).length;
    results.languages.count = Object.keys(results.languages.results).length;
    results.unknown.count = Object.keys({ ...results.unknown.extensions, ...results.unknown.filenames }).length;
    // Return
    return results;
}
module.exports = analyse;
